# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import argparse
import logging
import os

import tqdm
from ai_models.model import Model

from anemoi.inference.runner import AUTOCAST
from anemoi.inference.runner import DefaultRunner

LOG = logging.getLogger(__name__)


class AIModelPlugin(Model):

    expver = None

    def add_model_args(self, parser):
        """To be implemented in subclasses to add model-specific arguments to the parser.

        Parameters
        ----------
        parser : argparse.ArgumentParser
            An instance of the parser to add arguments to.
        """
        pass

    def parse_model_args(self, args):
        parser = argparse.ArgumentParser()

        parser.add_argument("--checkpoint", required=not hasattr(self, "download_files"))
        parser.add_argument(
            "--autocast",
            type=str,
            choices=sorted(AUTOCAST.keys()),
        )

        self.add_model_args(parser)

        args = parser.parse_args(args)

        for k, v in vars(args).items():
            setattr(self, k, v)

        if args.checkpoint:
            self.runner = DefaultRunner(args.checkpoint)
        else:
            self.runner = DefaultRunner(os.path.join(self.assets, self.download_files[0]))

        return parser

    def run(self):
        if self.deterministic:
            self.torch_deterministic_mode()

        self.runner.run(
            input_fields=self.all_fields,
            lead_time=self.lead_time,
            start_datetime=self.start_datetime,
            device=self.device,
            output_callback=self._output,
            autocast=self.autocast,
            progress_callback=tqdm.tqdm,
        )

    def _output(self, *args, **kwargs):
        if "step" in kwargs or "endStep" in kwargs:
            return self.write(*args, **kwargs)

        return self.write_input_fields(*args, **kwargs)

    # Below are methods forwarded to the checkpoint

    @property
    def param_sfc(self):
        return self.runner.checkpoint.param_sfc

    @property
    def param_level_pl(self):
        return self.runner.checkpoint.param_level_pl

    @property
    def param_level_ml(self):
        return self.runner.checkpoint.param_level_ml

    @property
    def grid(self):
        return self.runner.checkpoint.grid

    @property
    def area(self):
        return self.runner.checkpoint.area

    @property
    def lagged(self):
        return self.runner.lagged


model = AIModelPlugin



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

from . import Metadata

LOG = logging.getLogger(__name__)


class Version_0_0_0(Metadata):
    """
    Reference for very old checkpoints
    Will not work and need to be updated
    """

    def __init__(self, metadata):
        super().__init__(metadata)

        FORCING_PARAMS = [
            "z",
            "lsm",
            "sdor",
            "slor",
            "cos_latitude",
            "cos_longitude",
            "sin_latitude",
            "sin_longitude",
            "cos_julian_day",
            "cos_local_time",
            "sin_julian_day",
            "sin_local_time",
            "insolation",
        ]

        indices = dict(
            forcing=self._index_of(FORCING_PARAMS),
            full=self._index_of(self.variables),
            diagnostic=[],
            prognostic=self._index_of(self.ordering),
        )

        config = dict(
            data_indices=dict(
                data=dict(
                    input=indices,
                    output=indices,
                ),
                model=dict(
                    input=indices,
                    output=indices,
                ),
            ),
            config=dict(
                data=dict(timestep=6, frequency=6),
                training=dict(
                    multistep_input=2,
                    precision="32",
                ),
            ),
        )

        self._metadata.update(config)

    def _index_of(self, names):
        return [self.variable_to_index[name] for name in names]

    def dump(self, indent=0):
        print("Version_0_0_0: Not implemented")

    # Input
    area = [90, 0, -90, 360]
    grid = [0.25, 0.25]
    param_sfc = [
        "z",
        "sp",
        "msl",
        "lsm",
        "sst",
        "sdor",
        "slor",
        "10u",
        "10v",
        "2t",
        "2d",
    ]
    param_level_pl = (
        ["q", "t", "u", "v", "w", "z"],
        [50, 100, 150, 200, 250, 300, 400, 500, 600, 700, 850, 925, 1000],
    )

    param_level_ml = ([], [])

    ordering = [
        "q_50",
        "q_100",
        "q_150",
        "q_200",
        "q_250",
        "q_300",
        "q_400",
        "q_500",
        "q_600",
        "q_700",
        "q_850",
        "q_925",
        "q_1000",
        "t_50",
        "t_100",
        "t_150",
        "t_200",
        "t_250",
        "t_300",
        "t_400",
        "t_500",
        "t_600",
        "t_700",
        "t_850",
        "t_925",
        "t_1000",
        "u_50",
        "u_100",
        "u_150",
        "u_200",
        "u_250",
        "u_300",
        "u_400",
        "u_500",
        "u_600",
        "u_700",
        "u_850",
        "u_925",
        "u_1000",
        "v_50",
        "v_100",
        "v_150",
        "v_200",
        "v_250",
        "v_300",
        "v_400",
        "v_500",
        "v_600",
        "v_700",
        "v_850",
        "v_925",
        "v_1000",
        "w_50",
        "w_100",
        "w_150",
        "w_200",
        "w_250",
        "w_300",
        "w_400",
        "w_500",
        "w_600",
        "w_700",
        "w_850",
        "w_925",
        "w_1000",
        "z_50",
        "z_100",
        "z_150",
        "z_200",
        "z_250",
        "z_300",
        "z_400",
        "z_500",
        "z_600",
        "z_700",
        "z_850",
        "z_925",
        "z_1000",
        "sp",
        "msl",
        "sst",
        "10u",
        "10v",
        "2t",
        "2d",
        "z",
        "lsm",
        "sdor",
        "slor",
    ]

    param_format = {"param_level": "{param}{levelist}"}

    computed_constants = [
        "cos_latitude",
        "cos_longitude",
        "sin_latitude",
        "sin_longitude",
    ]

    computed_forcing = [
        "cos_julian_day",
        "cos_local_time",
        "sin_julian_day",
        "sin_local_time",
        "insolation",
    ]

    @property
    def variables(self):
        return self.ordering + self.computed_constants + self.computed_forcing

    @property
    def variables_with_nans(self):
        return []

    ###########################################################################
    @property
    def order_by(self):
        return dict(
            valid_datetime="ascending",
            param_level=self.ordering,
            remapping={"param_level": "{param}_{levelist}"},
        )

    @property
    def select(self):
        return dict(
            param_level=self.variables,
            remapping={"param_level": "{param}_{levelist}"},
        )



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


def zarr_fill(metadata):
    metadata["variables"] = metadata["attrs"]["variables"]
    return metadata


def drop_fill(metadata):
    metadata["variables"] = [x for x in metadata["forward"]["variables"] if x not in metadata["drop"]]
    return metadata


def select_fill(metadata):
    metadata["variables"] = [x for x in metadata["forward"]["variables"] if x in metadata["select"]]
    return metadata


def rename_fill(metadata, select):

    rename = metadata["rename"]
    variables = metadata["forward"]["variables"]
    if select is not None:
        variables = [x for x in variables if x in select]

    metadata["variables"] = [rename.get(x, x) for x in variables]

    return metadata


def join_fill(metadata):
    variables = []
    for dataset in metadata["datasets"]:
        for k in dataset["variables"]:
            if k not in variables:
                variables.append(k)

    metadata["variables"] = variables
    return metadata


def patch(a, b):
    if "drop" in a:
        return drop_fill(
            {
                "action": "drop",
                "drop": a["drop"],
                "forward": zarr_fill({"action": "zarr", "attrs": b}),
            }
        )

    if "select" in a:
        return select_fill(
            {
                "action": "select",
                "select": a["select"],
                "forward": zarr_fill({"action": "zarr", "attrs": b}),
            }
        )

    if "rename" in a:
        return rename_fill(
            {
                "action": "rename",
                "rename": a["rename"],
                "forward": zarr_fill({"action": "zarr", "attrs": b}),
            },
            a.get("select"),
        )

    raise NotImplementedError(f"Cannot patch {a}")


def list_to_dict(datasets, config):

    # print(json.dumps(datasets, indent=4))
    # os._exit(1)

    arguments = config["dataloader"]["training"]
    assert "dataset" in arguments
    assert isinstance(arguments["dataset"], list)
    assert len(arguments["dataset"]) == len(datasets)

    patched = [patch(a, b) for a, b in zip(arguments["dataset"], datasets)]

    return {
        "specific": join_fill({"action": "join", "datasets": patched}),
        "version": "0.2.0",
        "arguments": arguments,
    }



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from functools import lru_cache

import pytest
from anemoi.datasets import open_dataset
from anemoi.datasets.data.stores import zarr_lookup

from anemoi.inference.checkpoint.metadata import Metadata

zarr_lookup = lru_cache(maxsize=None)(zarr_lookup)


def missing(*names):
    for name in names:
        if zarr_lookup(name, fail=False) is None:
            return True
    return False


def any_missing(dataset):
    if isinstance(dataset, dict):
        if "dataset" in dataset:
            if missing(dataset["dataset"]):
                return True

        if "datasets" in dataset:
            for v in dataset["datasets"]:
                if isinstance(v, str):
                    if missing(v):
                        return True
                elif any_missing(v):
                    return True

        for k, v in dataset.items():
            if any_missing(v):
                return True

        return False

    if isinstance(dataset, (list, tuple)):
        for v in dataset:
            if any_missing(v):
                return True

        return False

    return False


def standard_test(dataset):

    if any_missing(dataset):
        pytest.skip(f"Dataset {dataset} not available")

    ds = open_dataset(**dataset)
    md = Metadata.from_metadata({"version": "1.0.0", "dataset": ds.metadata()})
    md.dump(full=True)

    return ds, md


def test_cerra():
    standard_test({"dataset": "cerra-rr-an-oper-0001-mars-5p5km-1984-2008-6h-v1-smhi"})


def test_cutout():
    standard_test(
        {
            "adjust": "all",
            "cutout": [
                {"dataset": "cerra-rr-an-oper-0001-mars-5p5km-1984-2008-6h-v1-smhi", "thinning": 25},
                {"dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6"},
            ],
        }
    )


def test_grids():
    standard_test(
        {
            "adjust": "all",
            "grids": [
                {"dataset": "cerra-rr-an-oper-0001-mars-5p5km-1984-2008-6h-v1-smhi", "thinning": 25},
                {"dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6"},
            ],
        }
    )


def test_rename():
    standard_test(
        {
            "dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6",
            "rename": {"2t": "t2m"},
        }
    )


def test_drop():
    standard_test(
        {
            "dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6",
            "drop": {"2t"},
        }
    )


def test_select():
    standard_test(
        {
            "dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6",
            "select": {"2t"},
        }
    )


def test_subset():
    standard_test(
        {
            "dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6",
            "start": "2010-01-01",
            "end": "2010-01-02",
            "frequency": "6h",
        }
    )


def test_ensemble():
    standard_test(
        {
            "datasets": [
                "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6",
                "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6",
            ]
        }
    )


def test_join():
    standard_test(
        {
            "datasets": [
                {"dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6", "select": {"2t"}},
                {"dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6", "select": {"2d"}},
            ],
        }
    )


def test_concat():
    standard_test(
        {
            "datasets": [
                {"dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6", "end": 2010},
                {"dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6", "start": 2011},
            ],
        }
    )


def test_statistics():
    standard_test(
        {
            "dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6",
            "statistics": {"dataset": "aifs-ea-an-oper-0001-mars-o96-1979-2022-6h-v6"},
        }
    )


if __name__ == "__main__":
    test_drop()



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from functools import cached_property

from . import Metadata

LOG = logging.getLogger(__name__)


class Version_0_1_0(Metadata):
    def __init__(self, metadata):
        super().__init__(metadata)
        self.patch_metadata()

    @cached_property
    def _dataset(self):
        """
        Part of the metadata that refers to the zarr dataset
        """
        return self._metadata["dataset"]

    @cached_property
    def param_sfc(self):
        param_level_sfc = self._dataset["data_request"]["param_level"]["sfc"]
        param_step_sfc = self._dataset["data_request"]["param_step"]["sfc"]
        fc_param = [param for param, step in param_step_sfc if step != 0]
        return sorted([p for p in param_level_sfc if p not in fc_param])

    @cached_property
    def param_level_pl(self):
        if "pl" not in self._dataset["data_request"]["param_level"]:
            return [], []
        result = self._dataset["data_request"]["param_level"]["pl"]
        return sorted(set([r[0] for r in result])), sorted(set([r[1] for r in result]))

    @cached_property
    def param_level_ml(self):
        if "ml" not in self._dataset["data_request"]["param_level"]:
            return [], []
        result = self._dataset["data_request"]["param_level"]["ml"]
        return sorted(set([r[0] for r in result])), sorted(set([r[1] for r in result]))

    @cached_property
    def param_step(self):
        return []

    @cached_property
    def grid(self):
        return self._dataset["data_request"]["grid"]

    @cached_property
    def area(self):
        return self.rounded_area(self._dataset["data_request"]["area"])

    @cached_property
    def variables(self):
        return self._dataset["variables"]

    ###########################################################################

    def patch_metadata(self):
        drop = self._find("config", "dataloader", "training", "drop")
        if drop is not None:
            variables = self._find("dataset", "variables", default=[])
            pl = self._find("dataset", "data_request", "param_level", "pl", default=[])
            ml = self._find("dataset", "data_request", "param_level", "ml", default=[])
            sfc = self._find("dataset", "data_request", "param_level", "sfc", default=[])
            for var in drop:
                if var in variables:
                    variables.remove(var)
                if var in sfc:
                    sfc.remove(var)
                if "_" in var:
                    param, level = var.split("_")
                    level = int(level)
                    if [param, level] in pl:
                        pl.remove([param, level])
                    if [param, level] in ml:
                        ml.remove([param, level])

    @property
    def variables_with_nans(self):
        return []

    def dump(self, indent=0):
        print("Version_0_1_0: Not implemented")



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
import os
import zipfile
from functools import cached_property

from anemoi.utils.checkpoints import has_metadata
from anemoi.utils.checkpoints import load_metadata

from .metadata import Metadata

LOG = logging.getLogger(__name__)


class Checkpoint:
    def __init__(self, path):
        self.path = path
        self._metadata = None
        self._operational_config = None

    def __repr__(self):
        return self.path

    def __getattr__(self, name):
        if self._metadata is None:
            try:
                self._metadata = Metadata.from_metadata(load_metadata(self.path))
            except ValueError:
                if has_metadata(self.path):
                    raise
                self._metadata = Metadata.from_metadata(None)

        return getattr(self._metadata, name)

    def _checkpoint_metadata(self, name):
        return load_metadata(self.path, name)

    @cached_property
    def operational_config(self):
        try:
            result = load_metadata(self.path, "operational-config.json")
            LOG.info(f"Using operational configuration from checkpoint {self.path}")
            return result
        except ValueError:
            pass

        # Check for the operational-config.json file in the model directory
        path = os.path.join(os.path.dirname(self.path), "operational-config.json")
        if os.path.exists(path):
            with open(path, "r") as f:
                result = json.load(f)
                LOG.info(f"Using operational configuration from {path}")
                return result

        LOG.warning("No operational configuration found. Using default configuration.")
        return {}



#!/usr/bin/env python
# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import json
import logging
import os
import subprocess
from tempfile import TemporaryDirectory

import yaml

from . import Command

LOG = logging.getLogger(__name__)


class Metadata(Command):
    """Edit, remove, dump or load metadata from a checkpoint file."""

    def add_arguments(self, command_parser):
        from anemoi.utils.checkpoints import DEFAULT_NAME

        command_parser.add_argument("path", help="Path to the checkpoint.")

        group = command_parser.add_mutually_exclusive_group(required=True)

        group.add_argument(
            "--dump",
            action="store_true",
            help=(
                "Extract the metadata from the checkpoint and print it to the standard output"
                " or the file specified by ``--output``, in JSON or YAML format."
            ),
        )
        group.add_argument(
            "--load",
            action="store_true",
            help=(
                "Set the metadata in the checkpoint from the content"
                " of a file specified by the ``--input`` argument."
            ),
        )

        group.add_argument(
            "--edit",
            action="store_true",
            help=(
                "Edit the metadata in place, using the specified editor."
                " See the ``--editor`` argument for more information."
            ),
        )

        group.add_argument(
            "--remove",
            action="store_true",
            help="Remove the metadata from the checkpoint.",
        )

        command_parser.add_argument(
            "--name",
            default=DEFAULT_NAME,
            help="Name of metadata record to be used with the actions above.",
        )

        command_parser.add_argument(
            "--input",
            help="The output file name to be used by the ``--load`` option.",
        )

        command_parser.add_argument(
            "--output",
            help="The output file name to be used by the ``--dump`` option.",
        )

        command_parser.add_argument(
            "--editor",
            help="Editor to use for the ``--edit`` option. Default to ``$EDITOR`` if defined, else ``vi``.",
            default=os.environ.get("EDITOR", "vi"),
        )

        command_parser.add_argument(
            "--json",
            action="store_true",
            help="Use the JSON format with ``--dump`` and ``--edit``.",
        )

        command_parser.add_argument(
            "--yaml",
            action="store_true",
            help="Use the YAML format with ``--dump`` and ``--edit``.",
        )

    def run(self, args):
        if args.edit:
            return self.edit(args)

        if args.remove:
            return self.remove(args)

        if args.dump:
            return self.dump(args)

        if args.load:
            return self.load(args)

    def edit(self, args):

        from anemoi.utils.checkpoints import load_metadata
        from anemoi.utils.checkpoints import replace_metadata

        if args.json:
            ext = "json"
            dump = json.dump
            load = json.load
            kwargs = {"indent": 4, "sort_keys": True}
        else:
            ext = "yaml"
            dump = yaml.dump
            load = yaml.safe_load
            kwargs = {"default_flow_style": False}

        with TemporaryDirectory() as temp_dir:

            path = os.path.join(temp_dir, f"checkpoint.{ext}")
            metadata = load_metadata(args.path)

            with open(path, "w") as f:
                dump(metadata, f, **kwargs)

            subprocess.check_call([args.editor, path])

            with open(path) as f:
                edited = load(f)

            if edited != metadata:
                replace_metadata(args.path, edited)
            else:
                LOG.info("No changes made.")

    def remove(self, args):
        from anemoi.utils.checkpoints import remove_metadata

        remove_metadata(args.path, args.name)

    def dump(self, args):
        from anemoi.utils.checkpoints import load_metadata

        if args.output:
            file = open(args.output, "w")
        else:
            file = None

        metadata = load_metadata(args.path)

        if args.yaml:
            print(yaml.dump(metadata, indent=2, sort_keys=True), file=file)
            return

        if args.json or True:
            print(json.dumps(metadata, indent=4, sort_keys=True), file=file)
            return

    def load(self, args):
        from anemoi.utils.checkpoints import has_metadata
        from anemoi.utils.checkpoints import replace_metadata
        from anemoi.utils.checkpoints import save_metadata

        if args.input is None:
            raise ValueError("Please specify a value for --input")

        _, ext = os.path.splitext(args.input)
        if ext == ".json" or args.json:
            with open(args.input) as f:
                metadata = json.load(f)

        elif ext in (".yaml", ".yml") or args.yaml:
            with open(args.input) as f:
                metadata = yaml.safe_load(f)

        else:
            raise ValueError(f"Unknown file extension {ext}. Please specify --json or --yaml")

        if has_metadata(args.path, args.name):
            replace_metadata(args.path, metadata)
        else:
            save_metadata(args.path, metadata, args.name)


command = Metadata



#!/usr/bin/env python
# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from ..checkpoint import Checkpoint
from . import Command


class InspectCmd(Command):

    need_logging = False

    def add_arguments(self, command_parser):
        command_parser.add_argument("path", help="Path to the checkpoint.")
        command_parser.add_argument("--dump", action="store_true", help="Print internal information")

    def run(self, args):

        c = Checkpoint(args.path)

        if args.dump:
            c.dump(full=True)
            return

        c.dump(full=False)

        def _(f):
            try:
                return f()
            except Exception as e:
                return str(e)

        print("area:", _(lambda: c.area))
        print("computed_constants_mask:", _(lambda: c.computed_constants_mask))
        print("computed_constants:", _(lambda: c.computed_constants))
        print("computed_forcings_mask:", _(lambda: c.computed_forcings_mask))
        print("computed_forcings:", _(lambda: c.computed_forcings))
        print("constant_data_from_input_mask:", _(lambda: c.constant_data_from_input_mask))
        print("constants_from_input_mask:", _(lambda: c.constants_from_input_mask))
        print("constants_from_input:", _(lambda: c.constants_from_input))
        print("data_to_model:", _(lambda: c.data_to_model))
        print("diagnostic_output_mask:", _(lambda: c.diagnostic_output_mask))
        print("diagnostic_params:", _(lambda: c.diagnostic_params))
        print("grid:", _(lambda: c.grid))
        print("hour_steps:", _(lambda: c.hour_steps))
        print("imputable_variables:", _(lambda: c.imputable_variables))
        print("index_to_variable:", _(lambda: c.index_to_variable))
        print("model_to_data:", _(lambda: c.model_to_data))
        print("multi_step:", _(lambda: c.multi_step))
        print("num_input_features:", _(lambda: c.num_input_features))
        print("operational_config:", _(lambda: c.operational_config))
        print("order_by:", _(lambda: c.order_by))
        print("param_level_ml:", _(lambda: c.param_level_ml))
        print("param_level_pl:", _(lambda: c.param_level_pl))
        print("param_sfc:", _(lambda: c.param_sfc))
        print("precision:", _(lambda: c.precision))
        print("prognostic_data_input_mask:", _(lambda: c.prognostic_data_input_mask))
        print("prognostic_input_mask:", _(lambda: c.prognostic_input_mask))
        print("prognostic_output_mask:", _(lambda: c.prognostic_output_mask))
        print("prognostic_params:", _(lambda: c.prognostic_params))
        print("select:", _(lambda: c.select))
        print("variable_to_index:", _(lambda: c.variable_to_index))
        print("variables_with_nans:", _(lambda: c.variables_with_nans))
        print("variables:", _(lambda: c.variables))


command = InspectCmd



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from functools import cached_property

import numpy as np
import semantic_version

LOG = logging.getLogger(__name__)


def from_versions(checkpoint_version, dataset_version):
    from .version_0_0_0 import Version_0_0_0
    from .version_0_1_0 import Version_0_1_0
    from .version_0_2_0 import Version_0_2_0

    VERSIONS = {
        ("0.0.0", "0.0.0"): Version_0_0_0,
        ("1.0.0", "0.1.0"): Version_0_1_0,
        ("1.0.0", "0.2.0"): Version_0_2_0,
    }

    version = (
        semantic_version.Version.coerce(checkpoint_version),
        semantic_version.Version.coerce(dataset_version),
    )

    LOG.info("Versions: checkpoint=%s dataset=%s", *version)

    versions = {
        (
            semantic_version.Version.coerce(k[0]),
            semantic_version.Version.coerce(k[1]),
        ): v
        for k, v in VERSIONS.items()
    }

    candidate = None
    for v, klass in sorted(versions.items()):
        if version >= v:
            candidate = klass

    return candidate


class Metadata:
    def __init__(self, metadata):
        self._metadata = metadata

    def to_dict(self):
        return self._metadata

    @classmethod
    def from_metadata(cls, metadata):
        if metadata is None:
            metadata = dict(version="0.0.0", dataset=dict(version="0.0.0"))

        if isinstance(metadata["dataset"], list):
            from .patch import list_to_dict

            # Backward compatibility
            metadata["dataset"] = list_to_dict(metadata["dataset"], metadata["config"])

        if "arguments" not in metadata["dataset"]:
            metadata["dataset"]["version"] = "0.1.0"

        # When we changed from ecml_tools to anemoi-datasets, we went back in the
        # versionning
        if metadata["dataset"]["version"] in ("0.1.7", "0.1.8", "0.1.9"):
            metadata["dataset"]["version"] = "0.2.0"

        klass = from_versions(metadata["version"], metadata["dataset"]["version"])
        return klass(metadata)

    def _find(self, *keys, default=None):
        m = self._metadata
        for key in keys:
            m = m.get(key)
            if m is None:
                return default
        return m

    # Common properties

    ###########################################################################

    @cached_property
    def variable_to_index(self):
        return {v: i for i, v in enumerate(self.variables)}

    @cached_property
    def index_to_variable(self):
        return {i: v for i, v in enumerate(self.variables)}

    @cached_property
    def hour_steps(self):
        n = self._config_data.get("timestep", self._config_data["frequency"])
        try:
            return int(n)
        except ValueError:
            pass

        return int(n[:-1]) * {"h": 1, "d": 24}[n[-1]]

    ###########################################################################
    # Indices

    @cached_property
    def _indices(self):
        return self._metadata["data_indices"]

    @cached_property
    def num_input_features(self):
        return len(self._indices["model"]["input"]["full"])

    @cached_property
    def data_to_model(self):
        data = self._indices["data"]["input"]["full"]
        model = self._indices["model"]["input"]["full"]
        assert len(data) == len(model)
        return {i: j for i, j in zip(data, model)}

    @cached_property
    def model_to_data(self):
        data = self._indices["data"]["input"]["full"]
        model = self._indices["model"]["input"]["full"]
        assert len(data) == len(model)
        return {j: i for i, j in zip(data, model)}

    ###########################################################################
    @property
    def order_by(self):
        # order = self._dataset["order_by"]
        return dict(
            valid_datetime="ascending",
            param_level=self.variables,
            # ensemble=self.checkpoint.ordering('ensemble'),
            remapping={"param_level": "{param}_{levelist}"},
        )

    @property
    def select(self):
        # order = self._dataset["order_by"]
        return dict(
            # valid_datetime="ascending",
            param_level=self.variables,
            # ensemble=self.checkpoint.ordering('ensemble'),
            remapping={"param_level": "{param}_{levelist}"},
        )

    ###########################################################################
    @cached_property
    def _config_data(self):
        """Part of the metadata refers to the model configuration"""
        return self._metadata["config"]["data"]

    @cached_property
    def _config_training(self):
        """Part of the metadata refers to the model configuration"""
        return self._metadata["config"]["training"]

    ###########################################################################
    def _forcings(self, constants):
        forcing = self._indices["data"]["input"]["forcing"]
        data_mask = []
        model_mask = []
        names = []
        for f in forcing:
            if self.index_to_variable[f] in constants:
                data_mask.append(f)
                model_mask.append(self.data_to_model[f])
                names.append(self.index_to_variable[f])

        return data_mask, model_mask, names

    def _forcing_params(self):
        forcing = self._indices["data"]["input"]["forcing"]
        return [self.index_to_variable[f] for f in forcing]

    ###########################################################################

    @cached_property
    def _computed_constants(self):
        constants = [
            "cos_latitude",
            "cos_longitude",
            "sin_latitude",
            "sin_longitude",
        ]

        data_mask, model_mask, names = self._forcings(constants)

        LOG.debug("computed_constants data_mask: %s", data_mask)
        LOG.debug("computed_constants model_mask: %s", model_mask)
        LOG.info("Computed constants: %s", names)

        return data_mask, model_mask, names

    @property
    def computed_constants(self):
        return self._computed_constants[2]

    @property
    def computed_constants_mask(self):
        return self._computed_constants[1]

    ###########################################################################

    @cached_property
    def _computed_forcings(self):
        known = set(
            [
                "cos_julian_day",
                "cos_local_time",
                "sin_julian_day",
                "sin_local_time",
                "insolation",
            ]
        )

        print("FORCINGS", self._forcing_params())

        constants = set(self._forcing_params()) - set(self.constants_from_input) - set(self.computed_constants)

        if constants - known:
            LOG.warning(f"Unknown computed forcing {constants - known}")

        data_mask, model_mask, names = self._forcings(constants)

        LOG.debug("computed_forcing data_mask: %s", data_mask)
        LOG.debug("computed_forcing model_mask: %s", model_mask)
        LOG.info("Computed forcings: %s", names)

        return data_mask, model_mask, names

    @property
    def computed_forcings(self):
        return self._computed_forcings[2]

    @property
    def computed_forcings_mask(self):
        return self._computed_forcings[1]

    ###########################################################################

    @cached_property
    def _constants_from_input(self):
        """We assume that constants are single level variables"""
        params_sfc = set(self.param_sfc)
        constants = set(self._forcing_params()).intersection(params_sfc)

        data_mask, model_mask, names = self._forcings(constants)

        LOG.debug("constants_from_input: %s", data_mask)
        LOG.debug("constants_from_input: %s", model_mask)
        LOG.info("Constants from input: %s", names)

        return data_mask, model_mask, names

    @property
    def constants_from_input(self):
        return self._constants_from_input[2]

    @property
    def constants_from_input_mask(self):
        return self._constants_from_input[1]

    @property
    def constant_data_from_input_mask(self):
        return self._constants_from_input[0]

    ###########################################################################

    @cached_property
    def prognostic_input_mask(self):
        return np.array(self._indices["model"]["input"]["prognostic"])

    @cached_property
    def prognostic_data_input_mask(self):
        return np.array(self._indices["data"]["input"]["prognostic"])

    @cached_property
    def prognostic_output_mask(self):
        return np.array(self._indices["model"]["output"]["prognostic"])

    @cached_property
    def diagnostic_output_mask(self):
        return np.array(self._indices["model"]["output"]["diagnostic"])

    @cached_property
    def diagnostic_params(self):
        return [self.index_to_variable[i] for i in self._indices["data"]["input"]["diagnostic"]]

    @cached_property
    def prognostic_params(self):
        return [self.index_to_variable[i] for i in self._indices["data"]["input"]["prognostic"]]

    ###########################################################################
    @cached_property
    def precision(self):
        return self._config_training["precision"]

    @cached_property
    def multi_step(self):
        return self._config_training["multistep_input"]

    @cached_property
    def imputable_variables(self):
        return self.variables_with_nans

    def rounded_area(self, area):
        try:
            surface = (area[0] - area[2]) * (area[3] - area[1]) / 180 / 360
            if surface > 0.98:
                return [90, 0.0, -90, 360]
        except TypeError:
            pass
        return area

    def report_loading_error(self):
        import json

        if "provenance_training" not in self._metadata:
            return

        provenance_training = self._metadata["provenance_training"]

        LOG.error("Training provenance:\n%s", json.dumps(provenance_training, indent=2))



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
from functools import cached_property

from . import Metadata

LOG = logging.getLogger(__name__)


class DataRequest:
    def __init__(self, metadata):
        self.metadata = metadata

    @property
    def variables(self):
        return self.metadata["variables"]

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return name[:-7] if name.endswith("Request") else name

    def mars_request(self):

        def _as_list(v):
            if isinstance(v, list):
                return v
            return [v]

        def _as_string(r):
            r = {k: "/".join([str(x) for x in _as_list(v)]) for k, v in r.items() if v}
            return ",".join([f"{k}={v}" for k, v in r.items()])

        r = dict(grid=self.grid, area=self.area)
        yield _as_string(r)

        r = dict(param=_as_list(self.param_sfc))
        yield _as_string(r)

        param, pl = self.param_level_pl
        if param:
            r = dict(param=param, level=pl)
            yield _as_string(r)

        param, ml = self.param_level_ml
        if param:
            r = dict(param=param, level=ml)
            yield _as_string(r)

        nans = self.variables_with_nans
        if nans:
            r = dict(with_nans=nans)
            yield _as_string(r)

    def dump_content(self, full, indent):

        if full:
            print("{", f'"{self.__class__.__name__}"', ":")
            print(json.dumps(self.metadata, indent=2, default=str))
            print("}")
            return

        print()
        print(" " * indent, "-", self)
        if full:
            print(" " * indent, " ", json.dumps(self.metadata, indent=2, default=str))
        else:
            for n in self.mars_request():
                print(" " * indent, " ", n)

    @property
    def param_sfc(self):
        param_sfc = self.forward.param_sfc
        param_step_sfc = [p[0] for p in self.forward.param_step_sfc_pairs]
        return [p for p in param_sfc if p not in param_step_sfc]

    @property
    def param_level_pl(self):
        params = set([p[0] for p in self.param_level_pl_pairs])
        levels = set([p[1] for p in self.param_level_pl_pairs])
        return sorted(params), sorted(levels)

    @property
    def param_level_ml(self):
        params = set([p[0] for p in self.param_level_ml_pairs])
        levels = set([p[1] for p in self.param_level_ml_pairs])
        return sorted(params), sorted(levels)


class ZarrRequest(DataRequest):
    def __init__(self, metadata):
        super().__init__(metadata)
        self.attributes = metadata["attrs"]
        self.request = self.attributes["data_request"]

    @property
    def grid(self):
        return self.request["grid"]

    @property
    def area(self):
        return self.request["area"]

    @property
    def param_sfc(self):
        return self.request["param_level"].get("sfc", [])

    @property
    def param_level_pl_pairs(self):
        return self.request["param_level"].get("pl", [])

    @property
    def param_level_ml_pairs(self):
        return self.request["param_level"].get("ml", [])

    @property
    def param_step_sfc_pairs(self):
        return self.request["param_step"].get("sfc", [])

    @property
    def variables_with_nans(self):
        return sorted(self.attributes.get("variables_with_nans", []))

    def dump(self, full, indent=0):
        self.dump_content(full, indent)


class Forward(DataRequest):
    @cached_property
    def forward(self):
        return data_request(self.metadata["forward"])

    def __getattr__(self, name):
        return getattr(self.forward, name)

    def dump(self, full, indent=0):
        self.dump_content(full, indent)
        self.forward.dump(indent + 2, full)


class SubsetRequest(Forward):
    # Subset in time
    pass


class StatisticsRequest(Forward):
    pass


class RenameRequest(Forward):

    # Drop variables
    # No need to rename anything as self.metadata["variables"] is already up to date

    @property
    def variables_with_nans(self):
        rename = self.metadata["rename"]
        return sorted([rename.get(x, x) for x in self.forward.variables_with_nans])


class MultiRequest(Forward):
    def __init__(self, metadata):
        super().__init__(metadata)
        self.datasets = [data_request(d) for d in metadata["datasets"]]

    @cached_property
    def forward(self):
        return self.datasets[0]

    def dump(self, full, indent=0):
        self.dump_content(full, indent)
        for dataset in self.datasets:
            dataset.dump(indent + 2)


class JoinRequest(MultiRequest):
    """Join variables"""

    @property
    def param_sfc(self):
        result = []
        for dataset in self.datasets:
            for param in dataset.param_sfc:
                if param not in result:
                    result.append(param)
        return result

    @property
    def param_level_pl_pairs(self):
        result = []
        for dataset in self.datasets:
            for param in dataset.param_level_pl_pairs:
                if param not in result:
                    result.append(param)
        return result

    @property
    def param_level_ml(self):
        result = []
        for dataset in self.datasets_pairs:
            for param in dataset.param_level_ml_pairs:
                if param not in result:
                    result.append(param)
        return result

    @property
    def param_step_sfc_pairs(self):
        result = []
        for dataset in self.datasets:
            for param in dataset.param_step_sfc_pairs:
                if param not in result:
                    result.append(param)
        return result

    @property
    def variables_with_nans(self):
        result = set()
        for dataset in self.datasets:
            result.update(dataset.variables_with_nans)

        return sorted(result)


class ConcatRequest(MultiRequest):
    # Concat in time
    pass


class EnsembleRequest(MultiRequest):
    pass


class MultiGridRequest(MultiRequest):
    @property
    def grid(self):
        grids = [dataset.grid for dataset in self.datasets]
        return grids[0]
        raise NotImplementedError(";".join(str(g) for g in grids))

    @property
    def area(self):
        areas = [dataset.area for dataset in self.datasets]
        return areas[0]
        raise NotImplementedError(";".join(str(g) for g in areas))

    def mars_request(self):
        for d in self.datasets:
            yield from d.mars_request()


class GridsRequest(MultiGridRequest):
    pass


class CutoutRequest(MultiGridRequest):
    pass


class ThinningRequest(Forward):

    @property
    def grid(self):
        return f"thinning({self.forward.grid})"


class SelectRequest(Forward):
    # Select variables

    @property
    def param_sfc(self):
        return [x for x in self.forward.param_sfc if x in self.variables]

    @property
    def param_level_pl_pairs(self):
        return [x for x in self.forward.param_level_pl_pairs if f"{x[0]}_{x[1]}" in self.variables]

    @property
    def param_level_ml_pairs(self):
        return [x for x in self.forward.param_level_ml_pairs if f"{x[0]}_{x[1]}" in self.variables]

    @property
    def param_step_pairs(self):
        return [x for x in self.forward.param_step_pairs if x[0] in self.variables]

    @property
    def param_step_sfc_pairs(self):
        return [x for x in self.forward.param_step_sfc_pairs if x[0] in self.variables]

    @property
    def variables_with_nans(self):
        return [x for x in self.forward.variables_with_nans if x in self.variables]


class DropRequest(SelectRequest):

    # Drop variables
    # No need to drop anything as self.metadata["variables"] is already up to date

    @property
    def variables_with_nans(self):
        return [x for x in self.forward.variables_with_nans if x in self.variables]


def data_request(specific):
    action = specific.pop("action")
    action = action[0].upper() + action[1:].lower() + "Request"
    LOG.debug(f"DataRequest: {action}")
    return globals()[action](specific)


class Version_0_2_0(Metadata, Forward):
    def __init__(self, metadata):
        super().__init__(metadata)
        specific = metadata["dataset"]["specific"]
        self.forward = data_request(specific)

    @cached_property
    def area(self):
        return self.rounded_area(self.forward.area)

    #########################



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property

import numpy as np
import torch
from anemoi.utils.timer import Timer

from .checkpoint import Checkpoint

LOGGER = logging.getLogger(__name__)


AUTOCAST = {
    "16-mixed": torch.float16,
    "16": torch.float16,
    "32": torch.float32,
    "b16-mixed": torch.bfloat16,
    "b16": torch.bfloat16,
    "bf16-mixed": torch.bfloat16,
    "bf16": torch.bfloat16,
    "bfloat16": torch.bfloat16,
    "f16": torch.float16,
    "f32": torch.float32,
    "float16": torch.float16,
    "float32": torch.float32,
}


def forcing_and_constants(source, date, param):
    import climetlab as cml

    ds = cml.load_source(
        "constants",
        source,
        date=date,
        param=param,
    )

    assert len(ds) == len(param), (len(ds), len(param), date)

    return ds.to_numpy(dtype=np.float32)


def ignore(*args, **kwargs):
    pass


class Runner:
    """_summary_"""

    def __init__(self, checkpoint):
        self.checkpoint = Checkpoint(checkpoint)

    def run(
        self,
        *,
        input_fields,
        lead_time,
        device,
        start_datetime=None,
        output_callback=ignore,
        autocast=None,
        progress_callback=ignore,
    ):
        """_summary_

        Parameters
        ----------
        input_fields : _type_
            _description_
        lead_time : _type_
            _description_
        device : _type_
            _description_
        start_datetime : _type_, optional
            _description_, by default None
        output_callback : _type_, optional
            _description_, by default ignore
        autocast : _type_, optional
            _description_, by default None
        progress_callback : _type_, optional
            _description_, by default ignore

        Returns
        -------
        _type_
            _description_

        Raises
        ------
        RuntimeError
            _description_
        ValueError
            _description_
        """

        if autocast is None:
            autocast = self.checkpoint.precision

        if autocast is None:
            LOGGER.warning("No autocast given, using float16")
            autocast = "16"

        autocast = AUTOCAST[autocast]

        input_fields = input_fields.sel(**self.checkpoint.select)
        input_fields = input_fields.order_by(**self.checkpoint.order_by)

        number_of_grid_points = len(input_fields[0].grid_points()[0])

        LOGGER.info("Loading input: %d fields (lagged=%d)", len(input_fields), len(self.lagged))

        input_fields_numpy = input_fields.to_numpy(dtype=np.float32, reshape=False)
        print(input_fields_numpy.shape)

        input_fields_numpy = input_fields_numpy.reshape(
            len(self.lagged),
            len(input_fields) // len(self.lagged),
            number_of_grid_points,
        )  # nlags, nparams, ngrid

        # Used to check if we cover the whole input, with no overlaps
        check = np.full(self.checkpoint.num_input_features, fill_value=False, dtype=np.bool_)

        kinds = np.full(self.checkpoint.num_input_features, fill_value="?", dtype=np.character)

        inputs = np.full(self.checkpoint.num_input_features, fill_value=False, dtype=np.bool_)

        # E.g cos_latitude
        computed_constant_mask = self.checkpoint.computed_constants_mask
        assert not np.any(check[computed_constant_mask]), check
        check[computed_constant_mask] = True
        kinds[computed_constant_mask] = "C"

        # E.g. lsm, orography
        constant_from_input_mask = self.checkpoint.constants_from_input_mask
        assert not np.any(check[constant_from_input_mask]), check
        check[constant_from_input_mask] = True
        kinds[constant_from_input_mask] = "K"
        inputs[constant_from_input_mask] = True

        # e.g. isolation
        computed_forcing_mask = self.checkpoint.computed_forcings_mask
        assert not np.any(check[computed_forcing_mask]), check
        check[computed_forcing_mask] = True
        kinds[computed_forcing_mask] = "F"

        # e.g 2t, 10u, 10v
        prognostic_input_mask = self.checkpoint.prognostic_input_mask
        assert not np.any(check[prognostic_input_mask]), check
        check[prognostic_input_mask] = True
        kinds[prognostic_input_mask] = "P"
        inputs[prognostic_input_mask] = True

        #
        if not np.all(check):
            for i, c in enumerate(check):
                if not c:
                    LOGGER.error(
                        "Missing %s %s %s",
                        i,
                        self.checkpoint.model_to_data[i],
                        self.checkpoint.index_to_variable[self.checkpoint.model_to_data[i]],
                    )
            raise RuntimeError("Missing fields")

        prognostic_data_from_retrieved_fields_mask = []
        constant_data_from_retrieved_fields_mask = []

        MARS = {False: " ", True: "X"}

        retrieved_fields_index = 0
        for i, c in enumerate(check):
            if inputs[i]:
                assert kinds[i].decode() in ("P", "K")
                if kinds[i].decode() == "P":
                    prognostic_data_from_retrieved_fields_mask.append(retrieved_fields_index)
                else:
                    constant_data_from_retrieved_fields_mask.append(retrieved_fields_index)
                retrieved_fields_index += 1

            if hasattr(self, "verbose") and self.verbose:
                print(
                    "{:4d} {:1s} {} {:4d} {:10s}".format(
                        i,
                        kinds[i].decode(),
                        MARS[inputs[i]],
                        self.checkpoint.model_to_data[i],
                        self.checkpoint.index_to_variable[self.checkpoint.model_to_data[i]],
                    )
                )

        prognostic_data_from_retrieved_fields_mask = np.array(prognostic_data_from_retrieved_fields_mask)
        constant_data_from_retrieved_fields_mask = np.array(constant_data_from_retrieved_fields_mask)

        # Build the input tensor

        input_tensor_numpy = np.full(
            shape=(
                len(self.lagged),
                self.checkpoint.num_input_features,
                number_of_grid_points,
            ),
            fill_value=np.nan,
            dtype=np.float32,
        )  # nlags, nparams, ngrid

        # Check that the computed constant mask and the constant from input mask are disjoint
        # assert np.amax(prognostic_input_mask) < np.amin(constant_from_input_mask)

        input_tensor_numpy[:, prognostic_input_mask] = input_fields_numpy[:, prognostic_data_from_retrieved_fields_mask]

        input_tensor_numpy[:, constant_from_input_mask] = input_fields_numpy[
            :, constant_data_from_retrieved_fields_mask
        ]

        if start_datetime is None:
            start_datetime = input_fields.order_by(valid_datetime="ascending")[-1].datetime()

        constants = forcing_and_constants(
            source=input_fields[:1],
            param=self.checkpoint.computed_constants,
            date=start_datetime,
        )

        for i in range(len(self.lagged)):
            input_tensor_numpy[i, computed_constant_mask] = constants

        for i in range(len(self.lagged)):
            forcings = forcing_and_constants(
                source=input_fields[:1],
                param=self.checkpoint.computed_forcings,
                date=start_datetime + datetime.timedelta(hours=self.lagged[i]),
            )
            input_tensor_numpy[i, computed_forcing_mask] = forcings

        LOGGER.info("Input tensor shape: %s", input_tensor_numpy.shape)

        imputable_variables = self.checkpoint.imputable_variables
        can_be_missing = set()
        # check for NaNs
        for i in range(input_tensor_numpy.shape[1]):
            name = self.checkpoint.index_to_variable[self.checkpoint.model_to_data[i]]
            has_missing = np.isnan(input_tensor_numpy[:, i, :]).any()
            is_imputable = name in imputable_variables
            if has_missing:
                can_be_missing.add(name)
                if not is_imputable:
                    model_index = self.checkpoint.model_to_data[i]
                    LOGGER.error(
                        "No imputation specified for NaNs in %s (%s %s)",
                        name,
                        i,
                        model_index,
                    )
                    raise ValueError(f"Field '{name}' has NaNs and is not marked as imputable")

        with Timer(f"Loading {self.checkpoint}"):
            try:
                model = torch.load(
                    self.checkpoint.path,
                    map_location=device,
                ).to(device)
            except Exception:
                self.checkpoint.report_loading_error()
                raise

        model.eval()

        torch.set_grad_enabled(False)

        input_tensor_torch = torch.from_numpy(
            np.swapaxes(
                input_tensor_numpy,
                -2,
                -1,
            )[np.newaxis, ...]
        ).to(device)

        prognostic_output_mask = self.checkpoint.prognostic_output_mask
        diagnostic_output_mask = self.checkpoint.diagnostic_output_mask

        LOGGER.info("Using autocast %s", autocast)

        # Write dynamic fields
        def get_most_recent_datetime(input_fields):
            datetimes = [f.valid_datetime() for f in input_fields]
            latest = datetimes[-1]
            for d in datetimes:
                assert d <= latest, (datetimes, d, latest)
            return latest

        most_recent_datetime = get_most_recent_datetime(input_fields)
        reference_fields = [f for f in input_fields if f.valid_datetime() == most_recent_datetime]
        precip_template = reference_fields[self.checkpoint.variable_to_index["2t"]]

        accumulated_output = np.zeros(
            shape=(len(diagnostic_output_mask), number_of_grid_points),
            dtype=np.float32,
        )

        if self.checkpoint.diagnostic_params:
            output_callback(
                input_fields,
                self.checkpoint.diagnostic_params,
                precip_template,
                accumulated_output[0].shape,
            )
        else:
            output_callback(input_fields)

        prognostic_params = self.checkpoint.prognostic_params

        # with self.stepper(self.hour_steps) as stepper:

        for i in progress_callback(range(lead_time // self.hour_steps)):
            step = (i + 1) * self.hour_steps

            # Predict next state of atmosphere
            with torch.autocast(device_type=device, dtype=autocast):
                y_pred = model.predict_step(input_tensor_torch)

            # Detach tensor and squeeze
            output = np.squeeze(y_pred.cpu().numpy())

            prognostic_fields_numpy = output[:, prognostic_output_mask]
            if len(diagnostic_output_mask):
                diagnostic_fields_numpy = output[:, diagnostic_output_mask]

            for n, (m, param) in enumerate(zip(prognostic_data_from_retrieved_fields_mask, prognostic_params)):
                template = reference_fields[m]
                assert template.valid_datetime() == most_recent_datetime, (
                    template.valid_datetime(),
                    most_recent_datetime,
                )
                output_callback(
                    prognostic_fields_numpy[:, n],
                    template=template,
                    step=step,
                    check_nans=True,  # param in can_be_missing,
                )

            # Write diagnostic fields
            if len(diagnostic_output_mask):
                for n, param in enumerate(self.checkpoint.diagnostic_params):
                    accumulated_output[n] += np.maximum(0, diagnostic_fields_numpy[:, n])
                    assert precip_template.valid_datetime() == most_recent_datetime, (
                        precip_template.valid_datetime(),
                        most_recent_datetime,
                    )
                    output_callback(
                        accumulated_output[n],
                        stepType="accum",
                        template=precip_template,
                        startStep=0,
                        endStep=step,
                        param=param,
                        check_nans=True,  # param in can_be_missing,
                    )

            # Next step

            prognostic_fields = y_pred[..., prognostic_output_mask]

            # Compute new forcing

            forcing = forcing_and_constants(
                source=input_fields[:1],
                param=self.checkpoint.computed_forcings,
                date=start_datetime + datetime.timedelta(hours=step),
            )
            forcing = np.swapaxes(forcing[np.newaxis, np.newaxis, ...], -2, -1)
            forcing = torch.from_numpy(forcing).to(device)

            # Update dynamic tensor for next iteration
            input_tensor_torch = input_tensor_torch.roll(-1, dims=1)
            input_tensor_torch[:, -1, :, prognostic_input_mask] = prognostic_fields
            input_tensor_torch[:, -1, :, computed_forcing_mask] = forcing

            # progress_callback(i)

    @cached_property
    def hour_steps(self):
        return self.checkpoint.hour_steps

    @cached_property
    def lagged(self):
        result = list(range(0, self.checkpoint.multi_step))
        result = [-s * self.hour_steps for s in result]
        return sorted(result)


class DefaultRunner(Runner):
    """_summary_

    Parameters
    ----------
    Runner : _type_
        _description_
    """

    pass
